## camera.pro : globals and support macros : -*- sh -*-
## this file contains the tasks for running the camera analysis stage
## these tasks use the book camPendingExp

# test for required global variables
check.globals

book init camPendingExp
book init camPendingCleanup

macro camera.status
  book listbook camPendingExp
  book listbook camPendingCleanup
end

macro camera.reset
  book init camPendingExp
  book init camPendingCleanup
end

macro camera.on
  task camera.exp.load
    active true
  end
  task camera.exp.run
    active true
  end
end

macro camera.off
  task camera.exp.load
    active false
  end
  task camera.exp.run
    active false
  end
end

macro camera.cleanup.on
  task camera.cleanup.load
    active true
  end
  task camera.cleanup.run
    active true
  end
end

macro camera.cleanup.off
  task camera.cleanup.load
    active false
  end
  task camera.cleanup.run
    active false
  end
end

# this variable will cycle through the known database names
$camera_DB = 0

# select images ready for camera analysis
# new entries are added to camPendingExp
# skip already-present entries
task	       camera.exp.load
  host         local

  periods      -poll $LOADPOLL
  periods      -exec $LOADEXEC
  periods      -timeout 30
  npending     1

  stdout NULL
  stderr $LOGDIR/camera.exp.log

  task.exec
    $run = camtool -pendingexp
    if ($DB:n == 0)
      option DEFAULT
    else
      # save the DB name for the exit tasks
      option $DB:$camera_DB
      $run = $run -dbname $DB:$camera_DB
      $camera_DB ++
      if ($camera_DB >= $DB:n) set camera_DB = 0
    end
    add_poll_args run
    command $run
  end

  # success
  task.exit    0
    # convert 'stdout' to book format
    ipptool2book stdout camPendingExp -key cam_id -uniq -setword dbname $options:0 -setword pantaskState INIT
    if ($VERBOSE > 2)
      book listbook camPendingExp
    end

    # delete existing entries in the appropriate pantaskStates
    process_cleanup camPendingExp
  end

  # default exit status
  task.exit    default
    showcommand failure
  end

  # operation times out?
  task.exit    timeout
    showcommand timeout
  end
end

# run the cameraexposure script on pending images
task	       camera.exp.run
  periods      -poll $RUNPOLL
  periods      -exec $RUNEXEC
  periods      -timeout 10

  task.exec
    book npages camPendingExp -var N
    if ($N == 0) break
    if ($NETWORK == 0) break
    
    # look for new images in camPendingExp (pantaskState == INIT)
    book getpage camPendingExp 0 -var pageName -key pantaskState INIT
    if ("$pageName" == "NULL") break

    book setword camPendingExp $pageName pantaskState RUN
    book getword camPendingExp $pageName camera -var CAMERA
    book getword camPendingExp $pageName exp_tag -var EXP_TAG
    book getword camPendingExp $pageName cam_id -var CAM_ID
    book getword camPendingExp $pageName workdir -var WORKDIR_TEMPLATE
    book getword camPendingExp $pageName dvodb  -var DVODB
    book getword camPendingExp $pageName dbname -var DBNAME
    book getword camPendingExp $pageName reduction -var REDUCTION
    book getword camPendingExp $pageName state -var RUN_STATE

    # specify choice of remote host based on camera and chip (class_id)
    set.host.for.camera $CAMERA FPA

    # set the WORKDIR variable
    set.workdir.by.camera $CAMERA FPA $WORKDIR_TEMPLATE $default_host WORKDIR

    # notes on how this works:
    # -- raw workdir examples:
    # file://data/@HOST@.0/gpc1/20080130
    # neb:///@HOST@-vol0/gpc1/20080130 (need to supply volname?, or are we re-defining this each time?)
    # -- out workdir examples:
    # file://data/ipp004.0/gpc1/20080130
    # neb:///ipp004-vol0/gpc1/20080130

    ## generate outroot specific to this exposure (& chip)
    sprintf outroot "%s/%s/%s.cm.%s" $WORKDIR $EXP_TAG $EXP_TAG $CAM_ID

    stdout $LOGDIR/camera.exp.log
    stderr $LOGDIR/camera.exp.log

    $run = camera_exp.pl --exp_tag $EXP_TAG --cam_id $CAM_ID --camera $CAMERA --outroot $outroot --redirect-output --run-state $RUN_STATE
    if ("$REDUCTION" != "NULL")
      $run = $run --reduction $REDUCTION
    end
    if ("$DVODB" != "NULL")
      $run = $run --dvodb $DVODB
    end
    add_standard_args run

    # save the pageName for future reference below
    options $pageName

    # create the command line
    if ($VERBOSE > 1)
      echo command $run
    end
    command $run
  end

  # success
  task.exit default
    process_exit camPendingExp $options:0 $JOB_STATUS
  end

  # operation times out?
  task.exit    timeout
    showcommand timeout
    book setword camPendingExp $options:0 pantaskState TIMEOUT
  end
end

# this variable will cycle through the known database names
$camera_cleanup_DB = 0

# select images ready for cam analysis
# new entries are added to camPendingImfile
# skip already-present entries
task	       camera.cleanup.load
  host         local

  periods      -poll $LOADPOLL
  periods      -exec $LOADEXEC
  periods      -timeout 30
  npending     1
  active       false

  stdout NULL
  stderr $LOGDIR/camera.cleanup.log

  task.exec
    $run = camtool -pendingcleanuprun
    if ($DB:n == 0)
      option DEFAULT
    else
      # save the DB name for the exit tasks
      option $DB:$camera_cleanup_DB
      $run = $run -dbname $DB:$camera_cleanup_DB
      $camera_cleanup_DB ++
      if ($camera_cleanup_DB >= $DB:n) set camera_cleanup_DB = 0
    end
    add_poll_args run
    command $run
  end

  # success
  task.exit    0
    # convert 'stdout' to book format
    ipptool2book stdout camPendingCleanup -key cam_id -uniq -setword dbname $options:0 -setword pantaskState INIT
    if ($VERBOSE > 2)
      book listbook camPendingCleanup
    end

    # delete existing entries in the appropriate pantaskStates
    process_cleanup camPendingCleanup
  end

  # locked list
  task.exit    default
    showcommand failure
  end

  # operation times out?
  task.exit    timeout
    showcommand timeout
  end
end

# run the ipp_cleanup.pl script on pending images
task	       camera.cleanup.run
  periods      -poll $RUNPOLL
  periods      -exec $RUNEXEC
  periods      -timeout 60
  active       false

  task.exec
    book npages camPendingCleanup -var N
    if ($N == 0) break
    if ($NETWORK == 0) break
    
    # look for new images in camPendingCleanup (pantaskState == INIT)
    book getpage camPendingCleanup 0 -var pageName -key pantaskState INIT
    if ("$pageName" == "NULL") break

    book setword camPendingCleanup $pageName pantaskState RUN
    book getword camPendingCleanup $pageName camera -var CAMERA
    book getword camPendingCleanup $pageName state  -var CLEANUP_MODE
    book getword camPendingCleanup $pageName cam_id -var CAM_ID
    book getword camPendingCleanup $pageName dbname -var DBNAME

    # specify choice of local or remote host based on camera and cam (class_id)
    set.host.for.camera $CAMERA FPA

    stdout $LOGDIR/camera.cleanup.log
    stderr $LOGDIR/camera.cleanup.log

    # XXX is everything listed here needed?
    $run = ipp_cleanup.pl --stage camera --stage_id $CAM_ID --camera $CAMERA --mode $CLEANUP_MODE
    add_standard_args run

    # save the pageName for future reference below
    options $pageName

    # create the command line
    if ($VERBOSE > 1)
      echo command $run
    end
    command $run
  end

  # default exit status
  task.exit    default
    process_exit camPendingCleanup $options:0 $JOB_STATUS
  end

  # operation timed out?
  task.exit    timeout
    showcommand timeout
    book setword camPendingCleanup $options:0 pantaskState TIMEOUT
  end
end
