## diff.pro : image difference analysis : -*- sh -*-

## This file contains panTasks definitions for performing the image differencing.
## After a difference (with associated diff_id) is defined, the difference is performed
## (tasks in diffSkyfile).

# test for required global variables
check.globals

### Initialise the books containing the tasks to do
book init diffSkyfile
book init diffCleanup

### Database lists
$diffSkycell_DB = 0
$diffCleanup_DB = 0

### Check status of diffing tasks
macro diff.status
  book listbook diffSkyfile
  book listbook diffCleanup
end

### Reset diffing tasks
macro diff.reset
  book init diffSkyfile
  book init diffCleanup
end

### Turn diffing tasks on
macro diff.on
  task diff.skycell.load
    active true
  end
  task diff.skycell.run
    active true
  end
end

### Turn diffing tasks off
macro diff.off
  task diff.skycell.load
    active false
  end
  task diff.skycell.run
    active false
  end
end

macro diff.cleanup.on
  task diff.cleanup.load
    active true
  end
  task diff.cleanup.run
    active true
  end
end

macro diff.cleanup.off
  task diff.cleanup.load
    active false
  end
  task diff.cleanup.run
    active false
  end
end


### Load tasks for doing the differences
### Tasks are loaded into diffSkyfile.
task	       diff.skycell.load
  host         local

  periods      -poll $LOADPOLL
  periods      -exec $LOADEXEC
  periods      -timeout 30
  npending     1

  stdout NULL
  stderr $LOGDIR/diff.skycell.log

  task.exec
    $run = difftool -todiffskyfile
    if ($DB:n == 0)
      option DEFAULT
    else
      # save the DB name for the exit tasks
      option $DB:$diffSkycell_DB
      $run = $run -dbname $DB:$diffSkycell_DB
      $diffSkycell_DB ++
      if ($diffSkycell_DB >= $DB:n) set diffSkycell_DB = 0
    end
    add_poll_args run
    command $run
  end

  # success
  task.exit    0
    # convert 'stdout' to book format
    ipptool2book stdout diffSkyfile -key diff_id:skycell_id -uniq -setword dbname $options:0 -setword pantaskState INIT
    if ($VERBOSE > 2)
      book listbook diffSkyfile
    end

    # delete existing entries in the appropriate pantaskStates
    process_cleanup diffSkyfile
  end

  # locked list
  task.exit    default
    showcommand failure
  end

  # operation times out?
  task.exit    timeout
    showcommand timeout
  end
end

### Run tasks for calculating the diff overlaps
### Tasks are taken from diffSkyfile.
task	       diff.skycell.run
  periods      -poll $RUNPOLL
  periods      -exec $RUNEXEC
  periods      -timeout 60

  task.exec
    book npages diffSkyfile -var N
    if ($N == 0) break
    if ($NETWORK == 0) break
    
    # look for new images in diffSkyfile (pantaskState == INIT)
    book getpage diffSkyfile 0 -var pageName -key pantaskState INIT
    if ("$pageName" == "NULL") break

    book setword diffSkyfile $pageName pantaskState RUN
    book getword diffSkyfile $pageName diff_id -var DIFF_ID
    book getword diffSkyfile $pageName diff_skyfile_id -var DIFF_SKYFILE_ID
    book getword diffSkyfile $pageName tess_id -var TESS_DIR
    book getword diffSkyfile $pageName skycell_id -var SKYCELL_ID
    book getword diffSkyfile $pageName camera -var CAMERA
    book getword diffSkyfile $pageName workdir -var WORKDIR_TEMPLATE
    book getword diffSkyfile $pageName dbname -var DBNAME

    # set the host and workdir based on the skycell hash
    set.host.for.skycell $SKYCELL_ID
    set.workdir.by.skycell $SKYCELL_ID $WORKDIR_TEMPLATE $default_host WORKDIR

    # XXX old code:
    # host anyhost
    # $WORKDIR = $WORKDIR_TEMPLATE

    basename $TESS_DIR -var TESS_ID
    sprintf outroot "%s/%s/%s/%s.%s.dif.%s" $WORKDIR $TESS_ID $SKYCELL_ID $TESS_ID $SKYCELL_ID $DIFF_ID

    stdout $LOGDIR/diff.skycell.log
    stderr $LOGDIR/diff.skycell.log

    $run = diff_skycell.pl --threads @MAX_THREADS@ --diff_id $DIFF_ID --skycell_id $SKYCELL_ID --outroot $outroot --redirect-output
    add_standard_args run

    # save the pageName for future reference below
    options $pageName

    # create the command line
    if ($VERBOSE > 1)
      echo command $run
    end
    command $run
  end

  # default exit status
  task.exit    default
    process_exit diffSkyfile $options:0 $JOB_STATUS
  end

  # operation timed out?
  task.exit    timeout
    showcommand timeout
    book setword diffSkyfile $options:0 pantaskState TIMEOUT
  end
end

# select images ready for diff analysis
# new entries are added to diffPendingImfile
# skip already-present entries
task	       diff.cleanup.load
  host         local

  periods      -poll $LOADPOLL
  periods      -exec $LOADEXEC
  periods      -timeout 30
  npending     1
  active       false

  stdout NULL
  stderr $LOGDIR/diff.cleanup.log

  task.exec
    $run = difftool -pendingcleanuprun
    if ($DB:n == 0)
      option DEFAULT
    else
      # save the DB name for the exit tasks
      option $DB:$diffCleanup_DB
      $run = $run -dbname $DB:$diffCleanup_DB
      $diffCleanup_DB ++
      if ($diffCleanup_DB >= $DB:n) set diffCleanup_DB = 0
    end
    add_poll_args run
    command $run
  end

  # success
  task.exit    0
    # convert 'stdout' to book format
    ipptool2book stdout diffCleanup -key diff_id:skycell_id -uniq -setword dbname $options:0 -setword pantaskState INIT
    if ($VERBOSE > 2)
      book listbook diffCleanup
    end

    # delete existing entries in the appropriate pantaskStates
    process_cleanup diffCleanup
  end

  # locked list
  task.exit    default
    showcommand failure
  end

  # operation times out?
  task.exit    timeout
    showcommand timeout
  end
end

# run the ipp_cleanup.pl script on pending images
task	       diff.cleanup.run
  periods      -poll $RUNPOLL
  periods      -exec $RUNEXEC
  periods      -timeout 60
  active       false

  task.exec
    book npages diffCleanup -var N
    if ($N == 0) break
    if ($NETWORK == 0) break
    
    # look for new images in diffCleanup (pantaskState == INIT)
    book getpage diffCleanup 0 -var pageName -key pantaskState INIT
    if ("$pageName" == "NULL") break

    book setword diffCleanup $pageName pantaskState RUN
    book getword diffCleanup $pageName camera -var CAMERA
    book getword diffCleanup $pageName state -var CLEANUP_MODE
    book getword diffCleanup $pageName diff_id -var DIFF_ID
    book getword diffCleanup $pageName dbname -var DBNAME

    # specify choice of local or remote host based on camera and diff (class_id)
    set.host.for.camera $CAMERA FPA

    stdout $LOGDIR/diff.cleanup.log
    stderr $LOGDIR/diff.cleanup.log

    # XXX is everything listed here needed?
    $run = ipp_cleanup.pl --stage diff --stage_id $DIFF_ID --camera $CAMERA --mode $CLEANUP_MODE
    add_standard_args run

    # save the pageName for future reference below
    options $pageName

    # create the command line
    if ($VERBOSE > 1)
      echo command $run
    end
    command $run
  end

  # default exit status
  task.exit    default
    process_exit diffCleanup $options:0 $JOB_STATUS
  end

  # operation timed out?
  task.exit    timeout
    showcommand timeout
    book setword diffCleanup $options:0 pantaskState TIMEOUT
  end
end
