/** @file ppSubSetMasks.c
 *
 *  @brief
 *
 *  @ingroup ppSub
 *
 *  @author IfA
 *  @version $Revision: 1.2 $ $Name: ipp-magic-v0-0 $
 *  @date $Date: 2009/02/06 01:37:17 $
 *  Copyright 2009 Institute for Astronomy, University of Hawaii
 */

#include "ppSub.h"

/** this function generates (if needed) and sets or updates the masks for both the input and
 * reference images.  this function also has the code for interpolation over bad pixels, but it
 * is currently if-def-ed out
 */

bool ppSubSetMasks (pmConfig *config, const pmFPAview *view) {

    psImageMaskType maskValue;
    psImageMaskType markValue;
    bool mdok = false;

    if (!pmConfigMaskSetBits(&maskValue, &markValue, config)) {
        psError(PS_ERR_UNKNOWN, false, "Unable to determine mask value.");
        return false;
    }

    // set the mask bits needed by psphot (in psphot recipe) 
    psphotSetMaskRecipe (config, maskValue, markValue);

    // Look up recipe values
    psMetadata *recipe = psMetadataLookupMetadata(&mdok, config->recipes, PPSUB_RECIPE); // Recipe for ppSim
    psAssert(recipe, "We checked this earlier, so it should be here.");

    psImageMaskType satValue = pmConfigMaskGet("SAT", config);
    psAssert (satValue, "SAT must be non-zero");

    psImageMaskType badValue = pmConfigMaskGet("BAD", config);
    psAssert (badValue, "BAD must be non-zero");

    // input images
    pmReadout *inRO = pmFPAfileThisReadout(config->files, view, "PPSUB.INPUT"); // Input readout
    pmReadout *refRO = pmFPAfileThisReadout(config->files, view, "PPSUB.REF"); // Reference readout

    psImage *input = inRO->image;       // Input image
    psImage *reference = refRO->image;  // Reference image
    PS_ASSERT_IMAGES_SIZE_EQUAL(input, reference, false);

    // XXX use psImageCopy below to avoid caring about image dimensions
    int numCols = input->numCols;
    int numRows = input->numRows;

    // Generate masks if they don't exist
    if (!inRO->mask) {
        if (psMetadataLookupBool(NULL, recipe, "MASK.GENERATE")) {
            pmReadoutSetMask(inRO, satValue, badValue);
        } else {
            inRO->mask = psImageAlloc(numCols, numRows, PS_TYPE_IMAGE_MASK);
            psImageInit(inRO->mask, 0);
        }
    }
    if (!refRO->mask) {
        if (psMetadataLookupBool(NULL, recipe, "MASK.GENERATE")) {
            pmReadoutSetMask(refRO, satValue, badValue);
        } else {
            refRO->mask = psImageAlloc(numCols, numRows, PS_TYPE_IMAGE_MASK);
            psImageInit(refRO->mask, 0);
        }
    }

    // Mask the NAN values
    if (!pmReadoutMaskNonfinite(inRO, satValue)) {
        psError(PS_ERR_UNKNOWN, false, "Unable to mask non-finite pixels in input.");
        return false;
    }
    if (!pmReadoutMaskNonfinite(refRO, satValue)) {
	psError(PS_ERR_UNKNOWN, false, "Unable to mask non-finite pixels in reference.");
	return false;
    }

#if (0)
    // Look up recipe values
    psMetadata *recipe = psMetadataLookupMetadata(NULL, config->recipes, PPSUB_RECIPE); // Recipe for ppSim
    psAssert(recipe, "We checked this earlier, so it should be here.");

    bool mdok = false;

    psString interpModeStr = psMetadataLookupStr(&mdok, recipe, "INTERPOLATION"); // Interpolation mode
    psImageInterpolateMode interpMode = psImageInterpolateModeFromString(interpModeStr); // Interp
    if (interpMode == PS_INTERPOLATE_NONE) {
        psError(PS_ERR_BAD_PARAMETER_VALUE, false, "Unknown interpolation mode: %s", interpModeStr);
        return false;
    }
    float poorFrac = psMetadataLookupF32(&mdok, recipe, "POOR.FRACTION"); // Fraction for "poor"

    psImageMaskType maskPoor = pmConfigMaskGet("POOR.WARP", config); // Bits to mask for poor pixels
    psImageMaskType maskBad = pmConfigMaskGet("BLANK", config); // Bits to mask for bad pixels

    // Interpolate over bad pixels, so the bad pixels don't explode
    if (!pmReadoutInterpolateBadPixels(inRO, maskVal, interpMode, poorFrac, maskPoor, maskBad)) {
	psError(PS_ERR_UNKNOWN, false, "Unable to interpolate bad pixels for input image.");
	return false;
    }
    if (!pmReadoutInterpolateBadPixels(refRO, maskVal, interpMode, poorFrac, maskPoor, maskBad)) {
	psError(PS_ERR_UNKNOWN, false, "Unable to interpolate bad pixels for reference image.");
	return false;
    }
    maskVal |= maskBad;
#endif

    return true;
}
